from .base import ExtractorBase
from .chatgpt import ChatGPTExtractor
from .extractor import Extractor
from .models import AnimeNameExtractResult, ResourceTitleExtractResult, VideoQuality
from .regex import RegexExtractor



from collections import defaultdict
from typing import Dict, List, Optional

from pydantic import BaseModel, Field, HttpUrl, field_validator, model_validator

from alist_mikananirss.alist import AlistDownloaderType
from alist_mikananirss.bot import PushPlusChannel

from ..utils import Singleton
from .config_loader import ConfigLoader


class AppConfig(BaseModel):
    # Common settings
    common_interval_time: int = Field(
        ge=0, description="Interval time must be non-negative"
    )
    common_proxies: Optional[Dict]

    # Alist settings
    alist_base_url: str
    alist_token: str
    alist_downloader: AlistDownloaderType
    alist_download_path: str

    # Mikan settings
    mikan_subscribe_url: List[str] = Field(min_length=1)
    mikan_regex_pattern: Dict
    mikan_filters: List

    # Notification settings
    notification_enable: bool
    notification_telegram_enable: bool
    notification_telegram_bot_token: str
    notification_telegram_user_id: str
    notification_pushplus_enable: bool
    notification_pushplus_token: str
    notification_pushplus_channel: PushPlusChannel
    notification_interval_time: int

    # Rename settings
    rename_enable: bool
    rename_chatgpt_api_key: str
    rename_chatgpt_base_url: str
    rename_chatgpt_model: str
    rename_format: str
    rename_remap_enable: bool
    rename_remap_cfg_path: str

    # Bot assistant settings
    bot_assistant_enable: bool
    bot_assistant_telegram_enable: bool
    bot_assistant_telegram_bot_token: str

    # Dev settings
    dev_log_level: str = Field(
        default="INFO", pattern="^(DEBUG|INFO|WARNING|ERROR|CRITICAL)$"
    )

    @field_validator("alist_base_url", "mikan_subscribe_url")
    @classmethod
    def validate_url(cls, url: str | List[str]) -> str | List[str]:
        if isinstance(url, list):
            for u in url:
                HttpUrl(u)
        else:
            HttpUrl(url)
        return url

    @model_validator(mode="after")
    def validate_notification_config(self) -> "AppConfig":
        if self.notification_enable:
            if not (
                self.notification_telegram_enable or self.notification_pushplus_enable
            ):
                raise ValueError("At least one notification method should be enabled")

            if self.notification_telegram_enable and not (
                self.notification_telegram_bot_token
                and self.notification_telegram_user_id
            ):
                raise ValueError("Telegram bot token and user id should be provided")

            if (
                self.notification_pushplus_enable
                and not self.notification_pushplus_token
            ):
                raise ValueError("Pushplus token should be provided")

        return self

    @field_validator("rename_format")
    @classmethod
    def validate_rename_format(cls, rename_format: str) -> str:

        if not rename_format:
            return rename_format

        all_key_test_data = {
            "name": "test",
            "season": 1,
            "episode": 1,
            "fansub": "fansub",
            "quality": "1080p",
            "language": "简体中文",
        }
        safe_dict = defaultdict(lambda: "undefined", all_key_test_data)
        res = rename_format.format_map(safe_dict)
        if "undefined" in res:
            unknown_keys = [
                key for key, value in safe_dict.items() if value == "undefined"
            ]
            raise ValueError(f"Error keys in rename format: {', '.join(unknown_keys)}")
        return rename_format

    @model_validator(mode="after")
    def validate_bot_assistant_config(self) -> "AppConfig":
        if self.bot_assistant_enable:
            if not self.bot_assistant_telegram_enable:
                raise ValueError("Telegram config should be provided")
            if not self.bot_assistant_telegram_bot_token:
                raise ValueError("Telegram bot token should be provided")
        return self

    def format_output_yaml(self) -> str:
        sections = {
            "common": {
                "interval_time": self.common_interval_time,
                "proxies": self.common_proxies,
            },
            "alist": {
                "base_url": self.alist_base_url,
                "token": "***" if self.alist_token else None,
                "downloader": self.alist_downloader,
                "download_path": self.alist_download_path,
            },
            "mikan": {
                "subscribe_url": self.mikan_subscribe_url,
                "regex_pattern": self.mikan_regex_pattern,
                "filters": self.mikan_filters,
            },
            "notification": {
                "telegram": {
                    "bot_token": "***" if self.notification_telegram_bot_token else "",
                    "user_id": "***" if self.notification_telegram_user_id else "",
                },
                "pushplus": {
                    "token": "***" if self.notification_pushplus_token else ""
                },
                "interval_time": self.notification_interval_time,
            },
            "rename": {
                "chatgpt": {
                    "api_key": "***" if self.rename_chatgpt_api_key else "",
                    "base_url": self.rename_chatgpt_base_url,
                    "model": self.rename_chatgpt_model,
                },
                "rename_format": self.rename_format,
                "remap": {
                    "enable": self.rename_remap_enable,
                    "cfg_path": self.rename_remap_cfg_path,
                },
            },
            "bot_assistant": {
                "telegram": {
                    "bot_token": "***" if self.bot_assistant_telegram_bot_token else "",
                },
            },
            "dev": {"log_level": self.dev_log_level},
        }

        def format_dict(d: dict, indent: int = 0) -> str:
            result = []
            for key, value in d.items():
                spaces = " " * indent
                if isinstance(value, dict):
                    result.append(f"{spaces}{key}:")
                    result.append(format_dict(value, indent + 2))
                elif isinstance(value, list):
                    result.append(f"{spaces}{key}:")
                    result.extend(f"{spaces}  - {item}" for item in value)
                else:
                    result.append(f"{spaces}{key}: {value}")
            return "\n".join(result)

        return format_dict(sections)


class ConfigManager(metaclass=Singleton):
    def __init__(self, config_path=None):
        if config_path:
            self.config: AppConfig = self.__load_config(config_path)
            self.config_path = config_path

    def __load_config(self, path):
        config_loader = ConfigLoader(path)
        # ----------common----------
        common_interval_time = config_loader.get("common.interval_time", 300)
        common_proxies = config_loader.get("common.proxies", None)
        # ----------alist----------
        alist_base_url = config_loader.get("alist.base_url")
        alist_token = config_loader.get("alist.token")
        alist_downloader = AlistDownloaderType(
            config_loader.get("alist.downloader", AlistDownloaderType.ARIA.value)
        )
        alist_download_path = config_loader.get("alist.download_path")
        # ----------mikan----------
        urls = config_loader.get("mikan.subscribe_url")
        if isinstance(urls, str):
            urls = [urls]
        mikan_subscribe_url = urls
        default_regex_pattern = {
            "简体": "(简体|简中|简日|CHS)",
            "繁体": "(繁体|繁中|繁日|CHT|Baha)",
            "1080p": "(X1080|1080P)",
            "非合集": "^(?!.*(\\d{2}-\\d{2}|合集)).*",
        }
        default_regex_pattern.update(config_loader.get("mikan.regex_pattern", {}))
        mikan_regex_pattern = default_regex_pattern
        mikan_filters = config_loader.get("mikan.filters", ["1080p", "非合集"])
        # ----------notification----------
        notification_enable = (
            False if not config_loader.get("notification", {}) else True
        )
        notification_telegram_enable = (
            False if not config_loader.get("notification.telegram", {}) else True
        )
        notification_telegram_bot_token = config_loader.get(
            "notification.telegram.bot_token", ""
        )
        notification_telegram_user_id = str(
            config_loader.get("notification.telegram.user_id", "")
        )
        notification_pushplus_enable = (
            False if not config_loader.get("notification.pushplus", {}) else True
        )
        notification_pushplus_token = config_loader.get(
            "notification.pushplus.token", ""
        )
        notification_pushplus_channel = PushPlusChannel(
            config_loader.get(
                "notification.pushplus.channel", PushPlusChannel.WECHAT.value
            )
        )
        notification_interval_time = config_loader.get(
            "notification.interval_time", 300
        )
        # ----------rename----------
        rename_enable = False if not config_loader.get("rename", {}) else True
        rename_chatgpt_api_key = config_loader.get("rename.chatgpt.api_key", "")
        rename_chatgpt_base_url = config_loader.get("rename.chatgpt.base_url", "")
        rename_chatgpt_model = config_loader.get("rename.chatgpt.model", "gpt-4o-mini")
        rename_format = config_loader.get(
            "rename.rename_format", "{name} S{season:02d}E{episode:02d}"
        )
        rename_remap_enable = config_loader.get("rename.remap.enable", False)
        rename_remap_cfg_path = config_loader.get("rename.remap.cfg_path", "remap.yaml")
        # ----------bot_assistant----------
        bot_assistant_enable = (
            False if not config_loader.get("bot_assistant", {}) else True
        )
        bot_assistant_telegram_enable = (
            False if not config_loader.get("bot_assistant.telegram", {}) else True
        )
        bot_assistant_telegram_bot_token = config_loader.get(
            "bot_assistant.telegram.bot_token", ""
        )
        # ----------dev----------
        dev_log_level = config_loader.get("dev.log_level", "INFO")

        return AppConfig(
            common_interval_time=common_interval_time,
            common_proxies=common_proxies,
            alist_base_url=alist_base_url,
            alist_token=alist_token,
            alist_downloader=alist_downloader,
            alist_download_path=alist_download_path,
            mikan_subscribe_url=mikan_subscribe_url,
            mikan_regex_pattern=mikan_regex_pattern,
            mikan_filters=mikan_filters,
            notification_enable=notification_enable,
            notification_telegram_enable=notification_telegram_enable,
            notification_telegram_bot_token=notification_telegram_bot_token,
            notification_telegram_user_id=notification_telegram_user_id,
            notification_pushplus_enable=notification_pushplus_enable,
            notification_pushplus_token=notification_pushplus_token,
            notification_pushplus_channel=notification_pushplus_channel,
            notification_interval_time=notification_interval_time,
            rename_enable=rename_enable,
            rename_chatgpt_api_key=rename_chatgpt_api_key,
            rename_chatgpt_base_url=rename_chatgpt_base_url,
            rename_chatgpt_model=rename_chatgpt_model,
            rename_format=rename_format,
            rename_remap_enable=rename_remap_enable,
            rename_remap_cfg_path=rename_remap_cfg_path,
            bot_assistant_enable=bot_assistant_enable,
            bot_assistant_telegram_enable=bot_assistant_telegram_enable,
            bot_assistant_telegram_bot_token=bot_assistant_telegram_bot_token,
            dev_log_level=dev_log_level,
        )

    def load_config(self, path):
        self.config = self.__load_config(path)
        self.config_path = path

    def reload_config(self):
        self.config = self.__load_config(self.config_path)

    def get_config(self):
        if not self.config:
            raise RuntimeError("Config not loaded")
        return self.config



import asyncio
import os
import time
from dataclasses import dataclass
from typing import Optional

from loguru import logger
from tenacity import (
    RetryError,
    retry,
    stop_after_attempt,
    wait_exponential,
)

from alist_mikananirss import utils
from alist_mikananirss.alist import Alist
from alist_mikananirss.alist.tasks import (
    AlistDownloadTask,
    AlistTask,
    AlistTaskStatus,
    AlistTaskType,
    AlistTransferTask,
)
from alist_mikananirss.common.database import SubscribeDatabase
from alist_mikananirss.websites.models import ResourceInfo

from ..utils import FixedSizeSet, Singleton
from .notification_sender import NotificationSender
from .renamer import AnimeRenamer


@dataclass
class AnimeDownloadTaskInfo:
    resource: ResourceInfo
    download_task: AlistDownloadTask
    transfer_task: Optional[AlistTransferTask] = None


class TaskMonitor:
    NORMAL_STATUS = [
        AlistTaskStatus.Pending,
        AlistTaskStatus.Running,
        AlistTaskStatus.StateBeforeRetry,
        AlistTaskStatus.StateWaitingRetry,
    ]
    STALL_THRESHOLD = 300  # 5min
    PROGRESS_THRESHOLD = 0.01  # 1%

    def __init__(
        self,
        alist: Alist,
        task: AlistTransferTask | AlistDownloadTask,
    ):
        self.alist = alist
        self.task = task
        self.last_progress = 0
        self.last_progress_time = time.time()

    @retry(
        stop=stop_after_attempt(5),
        wait=wait_exponential(multiplier=1, min=1, max=15),
    )
    async def _refresh(self):
        task_list = await self.alist.get_task_list(self.task.task_type)
        task = task_list.get_by_id(self.task.tid)
        if not task:
            raise RuntimeError("Can't find the task {self.task.tid}")
        self.task = task

    def _is_progress_stalled(self) -> bool:
        current_progress = self.task.progress
        current_time = time.time()

        time_elapsed = current_time - self.last_progress_time
        progress_change = current_progress - self.last_progress

        if (
            time_elapsed > self.STALL_THRESHOLD
            and progress_change < self.PROGRESS_THRESHOLD
            and self.task.status is AlistTaskStatus.Running
        ):
            return True

        if progress_change >= self.PROGRESS_THRESHOLD:
            self.last_progress = current_progress
            self.last_progress_time = current_time

        return False

    async def wait_finished(self) -> AlistTask:
        """Loop check task tatus until it finished or error

        Raises:
            TimeoutError: If the download progress remains unchanged for a long time, throw an exception
            RetryError: If the task status can't be refreshed, throw an exception

        Returns:
            AlistTask: The finished task
        """
        while True:
            await self._refresh()
            logger.debug(
                f"Checking {self.task} status: {self.task.status} progress: {self.task.progress:.2f}%"
            )
            if self.task.status not in self.NORMAL_STATUS:
                return self.task

            if self._is_progress_stalled():
                raise TimeoutError("Progress is too slow")

            await asyncio.sleep(1)


class DownloadManager(metaclass=Singleton):
    def __init__(
        self,
        alist_client: Alist,
        base_download_path: str,
        use_renamer: bool = False,
        need_notification: bool = False,
        db: SubscribeDatabase = None,
    ):
        self.alist_client = alist_client
        self.base_download_path = base_download_path
        self.use_renamer = use_renamer
        self.need_notification = need_notification
        self.uuid_set = FixedSizeSet()
        self.db = db

    @classmethod
    def initialize(
        cls,
        alist_client: Alist,
        base_download_path: str,
        use_renamer: bool = False,
        need_notification: bool = False,
        db: SubscribeDatabase = None,
    ) -> None:
        cls(
            alist_client=alist_client,
            base_download_path=base_download_path,
            use_renamer=use_renamer,
            need_notification=need_notification,
            db=db,
        )

    @classmethod
    async def add_download_tasks(cls, resources: list[ResourceInfo]):
        instance = cls()
        info_list: list[AnimeDownloadTaskInfo] = await instance.download(resources)
        for task_info in info_list:
            await instance.db.insert_resource_info(task_info.resource)
            asyncio.create_task(instance.monitor(task_info))

    @retry(
        stop=stop_after_attempt(7),
        wait=wait_exponential(multiplier=0.5, min=1, max=15),
        retry_error_callback=lambda _: None,
    )
    async def _find_transfer_task(
        self, download_task: AlistDownloadTask
    ) -> AlistTransferTask:
        try:
            transfer_task_list: list[AlistTransferTask] = (
                await self.alist_client.get_task_list(AlistTaskType.TRANSFER)
            )
        except Exception as e:
            logger.error(f"Error when getting transfer task list: {e}")
            raise
        for transfer_task in transfer_task_list:
            # 查找第一个，未被标记过的与下载任务的目标路径相同的视频文件传输任务作为下载任务对应的传输任务
            if (
                transfer_task.uuid not in self.uuid_set
                and utils.is_video(transfer_task.target_path)
                and transfer_task.status
                in [AlistTaskStatus.Pending, AlistTaskStatus.Running]
                and download_task.download_path in transfer_task.target_path
            ):
                self.uuid_set.add(transfer_task.uuid)
                logger.debug(f"Linked [{download_task.url}] to {transfer_task.uuid}")
                return transfer_task
        logger.warning(
            f"Can't find the transfer task of [{download_task.url}], retrying..."
        )
        raise TimeoutError("Transfer task not found")

    async def _wait_success(
        self, task: AnimeDownloadTaskInfo
    ) -> Optional[AnimeDownloadTaskInfo]:
        """monitor one task until it finished

        Args:
            task (AnimeDownloadTaskInfo): Task which need to be monitored

        Returns:
            AnimeDownloadTaskInfo | None: return the task info if download and transfer success, else return None
        """

        async def wait_task(task_obj: AlistTask):
            monitor = TaskMonitor(self.alist_client, task_obj)
            try:
                finished_task = await monitor.wait_finished()
                if finished_task.status == AlistTaskStatus.Succeeded:
                    return finished_task
                else:
                    logger.error(
                        f"Task {task_obj.tid} failed: {finished_task.error_msg}"
                    )
            except TimeoutError:
                await self.alist_client.cancel_task(task_obj)
                logger.error("Long time no progress, cancel the task: {task_obj}")
            except RetryError as e:
                logger.error(
                    f"Error to refresh task status: {e.last_attempt.exception()}"
                )
            return None

        success_download_task = await wait_task(task.download_task)
        if not success_download_task:
            return None

        transfer_task = await self._find_transfer_task(success_download_task)
        if not transfer_task:
            logger.error(
                f"Timeout to find the transfer task of {task.resource.resource_title}"
            )
            return None

        success_transfer_task = await wait_task(transfer_task)
        if not success_transfer_task:
            return None
        success_task_info = AnimeDownloadTaskInfo(
            resource=task.resource,
            download_task=success_download_task,
            transfer_task=success_transfer_task,
        )
        return success_task_info

    def _post_process(self, task: AnimeDownloadTaskInfo):
        "Something to do after download task success"
        logger.info(f"Download {task.resource.resource_title} success")
        if self.use_renamer:
            remote_filepath = task.transfer_task.target_path
            asyncio.create_task(AnimeRenamer.rename(remote_filepath, task.resource))
        if self.need_notification:
            asyncio.create_task(NotificationSender.add_resource(task.resource))

    async def monitor(self, task_info: AnimeDownloadTaskInfo):
        success_task = await self._wait_success(task_info)
        if success_task is not None:
            self._post_process(success_task)
        else:
            # 下载失败，删除数据库记录
            await self.db.delete_by_resource_title(task_info.resource.resource_title)

    def _build_download_path(self, resource: ResourceInfo) -> str:
        """build the download path based on the anime name and season
        The result looks like this: base_download_path/AnimeName/Season {season} \n
        if the anime name is not available, use the base download path

        Args:
            resource (ResourceInfo)

        Returns:
            str: new download path
        """
        download_path = self.base_download_path
        if resource.anime_name:
            # Replace illegal characters in the anime_name
            illegal_chars = ["\\", "/", ":", "*", "?", '"', "<", ">", "|"]
            anime_name = resource.anime_name
            for char in illegal_chars:
                anime_name = anime_name.replace(char, " ")
            download_path = os.path.join(download_path, anime_name)
            # Need to compare with None, or when seaon=0 will be ignored
            if resource.season is not None:
                download_path = os.path.join(download_path, f"Season {resource.season}")
        return download_path

    async def download(
        self, new_resources: list[ResourceInfo]
    ) -> list[AnimeDownloadTaskInfo]:
        """Create alist offline download task

        Args:
            new_resources (list[MikanAnimeResource]): resources list
            base_download_path (str): remote dir path

        Returns:
            list[AnimeDownloadTaskInfo]: Successful download task's info
        """
        # Generate a mapping of download paths to resources
        # facilitating batch creation of download tasks and reducing requests to the Alist API
        ## mapping of {download_path: [torrent_url]}
        path_urls: dict[str, list[str]] = {}
        for resource in new_resources:
            download_path = self._build_download_path(resource)
            path_urls.setdefault(download_path, []).append(resource.torrent_url)
        # start to request the Alist Download API
        task_list = []
        for download_path, urls in path_urls.items():
            try:
                task_list += await self.alist_client.add_offline_download_task(
                    download_path, urls
                )
                logger.info(
                    f"Start to download {len(urls)} resources to [{download_path}]"
                )
            except Exception as e:
                logger.error(f"Error when add offline download task: {e}")
                continue
        # Patch the download task with the resource information
        anime_task_list = [
            AnimeDownloadTaskInfo(
                resource=resource,
                download_task=task,
            )
            for task in task_list
            for resource in new_resources
            if resource.torrent_url == task.url
        ]
        return anime_task_list



import yaml


class ConfigLoader:
    """A class for loading and accessing YAML configuration files.

    This class provides methods to load a YAML config file and retrieve values
    using dot notation paths.

    Attributes:
        config_path (str): The path to the YAML configuration file.
        config (dict): The loaded configuration data.

    Example:
        >>> config_loader = ConfigLoader('config.yaml')
        >>> database_url = config_loader.get('database.url')
        >>> port = config_loader.get('server.port', default=8080)
    """

    _MISSING = object()

    def __init__(self, config_path):
        """Initializes the ConfigLoader with the given config file path.

        Args:
            config_path (str): The path to the YAML configuration file.
        """
        self.config_path = config_path
        self.config = self.load_config()

    def load_config(self):
        """Loads the YAML configuration file.

        Returns:
            dict: The loaded configuration data.

        Raises:
            yaml.YAMLError: If there's an error parsing the YAML file.
        """
        with open(self.config_path, "r", encoding="utf-8") as file:
            return yaml.safe_load(file)

    def get(self, path, default=_MISSING):
        """Retrieves a value from the configuration using a dot notation path.

        Args:
            path (str): The dot notation path to the desired configuration value.
            default: The default value to return if the path is not found.
                     If not provided, a KeyError will be raised.

        Returns:
            The value at the specified path in the configuration.

        Raises:
            KeyError: If the path is not found and no default value is provided.
        """
        keys = path.split(".")
        value = self.config
        for key in keys:
            if value.get(key) is None:
                if default is not self._MISSING:
                    return default
                else:
                    raise KeyError(
                        f"{path} is not found in config file {self.config_path}"
                    )
            value = value[key]
        return value



from typing import Optional

from loguru import logger
from openai import AsyncOpenAI

from alist_mikananirss.utils.tmdb import TMDBClient

from .base import ExtractorBase
from .models import (
    AnimeNameExtractResult,
    ResourceTitleExtractResult,
    TMDBSearchParam,
    TMDBTvInfo,
)


class ChatGPTExtractor(ExtractorBase):
    def __init__(self, api_key, base_url=None, model="gpt-4o-mini") -> None:
        self._api_key = api_key
        self._base_url = base_url
        self.model = model
        self._client = None

    @property
    def client(self):
        if self._client is None:
            self._client = AsyncOpenAI(api_key=self._api_key)
            if self._base_url:
                self._client.base_url = self._base_url
        return self._client

    async def analyse_anime_name(self, anime_name: str) -> AnimeNameExtractResult:
        response = await self.client.beta.chat.completions.parse(
            model=self.model,
            messages=[
                {
                    "role": "system",
                    "content": "You are an anime series resource categorization assistant. When given an anime series name, you need to parse out the original series name without season information, and the season information (default is Season 1)",
                },
                {"role": "user", "content": anime_name},
            ],
            response_format=AnimeNameExtractResult,
        )

        res = response.choices[0].message.parsed
        if res is None:
            raise ValueError(f"Failed to parse anime name: {anime_name} by GPT")
        logger.debug(f"Chatgpt analyse resource name: {anime_name} -> {res}")
        return res

    async def search_name_in_tmdb(
        self, resource_title: str, max_retry_times: int = 5
    ) -> Optional[TMDBTvInfo]:
        """Ask GPT use the resource title to search in TMDB and find the correct anime name

        Args:
            resource_title (str): The title of the resource

        Returns:
            TMDBTvInfo: The information of the tv series in TMDB, or None if not found
        """

        # 1. Ask GPT to parse the resource title and extract search keyword
        response = await self.client.beta.chat.completions.parse(
            model=self.model,
            messages=[
                {
                    "role": "system",
                    "content": "You are an anime series search assistant. You need to parse the filename of anime resources and extract search keywords from it. The keywords should be as concise as possible and should not include season information, to avoid missing relevant search results.",
                },
                {"role": "user", "content": resource_title},
            ],
            response_format=TMDBSearchParam,
        )
        search_param = response.choices[0].message.parsed
        if search_param is None:
            logger.error(f"Failed to parse resource title: {resource_title} by GPT")
            return None
        logger.debug(f"Search param: {search_param}")

        # 2. Use the keyword to search in TMDB
        self.tmdb_client = TMDBClient()
        search_results = await self.tmdb_client.search_tv(search_param.query)
        i = 0
        while i < max_retry_times and len(search_results) == 0:
            logger.warning(
                f"Unable to find anime name in TMDB, search param: {search_param}. Retry {i+1}/{max_retry_times}"
            )
            response = await self.client.beta.chat.completions.parse(
                model=self.model,
                messages=[
                    {
                        "role": "system",
                        "content": "You are an anime series search assistant. You excel at breaking down original anime titles into keywords containing core information, then entering these keywords into search engines to find corresponding results.",
                    },
                    {
                        "role": "user",
                        "content": f"No results found for: {search_param.query}. Please try a different keyword.",
                    },
                ],
                response_format=TMDBSearchParam,
            )
            search_param = response.choices[0].message.parsed
            if search_param is None:
                logger.warning(
                    f"Failed to parse resource title: {resource_title} by GPT"
                )
                continue
            search_results = await self.tmdb_client.search_tv(search_param.query)
            i += 1

        if len(search_results) == 0:
            logger.error("Unable to find anime name in TMDB")
            return None
        logger.debug(f"Search results: {search_results}")

        # 3. Ask GPT to find the correct anime in the search results
        response = await self.client.beta.chat.completions.parse(
            model=self.model,
            messages=[
                {
                    "role": "system",
                    "content": "You are an anime series search assistant. Now that the search part is completed, given a filename of an anime resource, you need to find the corresponding anime series from the search results based on the information contained in the filename",
                },
                {
                    "role": "user",
                    "content": f"resource_file_name: {resource_title}, search_results: {search_results}",
                },
            ],
            response_format=TMDBTvInfo,
        )
        message = response.choices[0].message
        tmdb_info = message.parsed
        if tmdb_info is None:
            logger.error(
                f"Failed to find the anime of {resource_title} in search result: {search_results} by GPT"
            )
            return None
        logger.debug(f"TMDB info: {tmdb_info}")

        return tmdb_info

    async def analyse_resource_title(
        self, resource_title: str, use_tmdb: bool = True
    ) -> ResourceTitleExtractResult:
        response = await self.client.beta.chat.completions.parse(
            model=self.model,
            messages=[
                {
                    "role": "system",
                    "content": "You are an anime resource categorization assistant. Given an anime resource name, you need to extract information that helps users categorize and organize the resource based on its name",
                },
                {
                    "role": "user",
                    "content": resource_title,
                },
            ],
            response_format=ResourceTitleExtractResult,
        )
        message = response.choices[0].message
        res = message.parsed
        if res is None:
            raise ValueError(f"Failed to parse resource title: {resource_title} by GPT")

        if use_tmdb:
            tmdb_info = await self.search_name_in_tmdb(resource_title)
            res.anime_name = tmdb_info.anime_name if tmdb_info else res.anime_name
        logger.debug(f"Chatgpt analyse resource name: {resource_title} -> {res}")
        return res



from unittest.mock import AsyncMock, patch

import pytest

from alist_mikananirss.alist.api import Alist
from alist_mikananirss.alist.tasks import (
    AlistDeletePolicy,
    AlistDownloaderType,
    AlistDownloadTask,
    AlistTaskStatus,
    AlistTaskType,
    AlistTransferTask,
)


@pytest.fixture
def alist():
    return Alist(
        base_url="https://example.com",
        token="test_token",
        downloader=AlistDownloaderType.ARIA,
    )


@pytest.mark.asyncio
async def test_alist_init(alist):
    assert alist.base_url == "https://example.com"
    assert alist.token == "test_token"
    assert alist.downloader == AlistDownloaderType.ARIA


@pytest.mark.asyncio
async def test_get_alist_ver(alist):
    with patch.object(alist, "_api_call", new_callable=AsyncMock) as mock_api_call:
        mock_api_call.return_value = {"version": "v2.0.0"}
        version = await alist.get_alist_ver()
        assert version == "2.0.0"
        mock_api_call.assert_called_once_with("GET", "/api/public/settings")


@pytest.mark.asyncio
async def test_add_offline_download_task(alist):
    with patch.object(alist, "_api_call", new_callable=AsyncMock) as mock_api_call:
        mock_api_call.return_value = {
            "tasks": [
                {
                    "error": "",
                    "id": "lwYyV7TlpdS06fiflUmBH",
                    "name": "download magnet:?xt=xxx to (/Local/颂乐人偶/Season 1)",
                    "progress": 0.0,
                    "state": 1,
                    "status": "offline download waiting",
                }
            ]
        }
        tasks = await alist.add_offline_download_task(
            "/Local/颂乐人偶/Season 1", ["magnet:?xt=xxx"]
        )
        assert isinstance(tasks[0], AlistDownloadTask)
        mock_api_call.assert_called_once_with(
            "POST",
            "api/fs/add_offline_download",
            json={
                "delete_policy": AlistDeletePolicy.DeleteAlways.value,
                "path": "/Local/颂乐人偶/Season 1",
                "urls": ["magnet:?xt=xxx"],
                "tool": alist.downloader.value,
            },
        )


@pytest.mark.asyncio
async def test_list_dir(alist):
    with patch.object(alist, "_api_call", new_callable=AsyncMock) as mock_api_call:
        mock_api_call.return_value = {
            "content": [{"name": "file1.txt"}, {"name": "file2.txt"}]
        }
        files = await alist.list_dir("/test/path")
        assert files == ["file1.txt", "file2.txt"]


@pytest.mark.asyncio
async def test_cancel_task(alist):
    with patch.object(alist, "_api_call", new_callable=AsyncMock) as mock_api_call:
        json_data = {
            "error": "",
            "id": "tid1",
            "name": "download xxx to (xxx)",
            "progress": 50.0,
            "state": 1,
            "status": "active",
        }
        task = AlistDownloadTask.from_json(json_data)
        result = await alist.cancel_task(task)
        assert result is True
        mock_api_call.assert_called_once_with(
            "POST", f"/api/task/offline_download/cancel?tid={json_data['id']}"
        )


@pytest.mark.asyncio
async def test_rename(alist):
    with patch.object(alist, "_api_call", new_callable=AsyncMock) as mock_api_call:
        await alist.rename("/old/path/file.txt", "new_file.txt")
        mock_api_call.assert_called_once_with(
            "POST",
            "api/fs/rename",
            json={"path": "/old/path/file.txt", "name": "new_file.txt"},
        )


def test_dl_task_extract():
    task = AlistDownloadTask.from_json(
        {
            "error": "",
            "id": "lwYyV7TlpdS06fiflUmBH",
            "name": "download magnet:?xt=xxx to (/Local/颂乐人偶/Season 1)",
            "progress": 100,
            "state": 2,
            "status": "offline download completed, maybe transferring",
        }
    )

    assert task.url == "magnet:?xt=xxx"
    assert task.download_path == "/Local/颂乐人偶/Season 1"
    assert task.task_type == AlistTaskType.DOWNLOAD

    # seeding situation
    task2 = AlistDownloadTask.from_json(
        {
            "error": "",
            "id": "lwYyV7TlpdS06fiflUmBH",
            "name": "download magnet:?xt=xxx to (/Local/颂乐人偶/Season 1)",
            "progress": 100,
            "state": 1,
            "status": "offline download completed, waiting for seeding",
        }
    )
    assert task2.status == AlistTaskStatus.Succeeded


def test_tf_task_extract():
    task = AlistTransferTask.from_json(
        {
            "error": "",
            "id": "i-T7dWBTgh_9bohMpRlcA",
            "name": "transfer [/Local](/opt/alist/data/temp/qBittorrent/d82ddec1-08f6-4894-b7ed-d9c9f25dc4db/test.mp4) to [/Google](/Debug/test/Season 1)",
            "progress": 100.0000041135186,
            "state": 2,
            "status": "transferring",
        }
    )

    assert task.uuid == "d82ddec1-08f6-4894-b7ed-d9c9f25dc4db"
    assert task.target_path == "/Google/Debug/test/Season 1/test.mp4"
    assert task.task_type == AlistTaskType.TRANSFER
